/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdint.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h> 
#include <dirent.h> 
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>

/* 1. Type Definition                                                 */
/**********************************************************************/
typedef uint32_t INST_TYPE;
typedef uint64_t ADDR_TYPE;
typedef uint64_t DATA_TYPE;


/* 2. Constant Definition                                             */
/**********************************************************************/

#define VER "SimAlpha release version 1.1.0 2003-06-13"

#define DEF_FILE_NAME    "aout.txt"

#define BLOCK_SIZE       0x00002000    /* 8KB page size        */
#define BLOCK_TABLE_SIZE 0x00080000    /* size of page table   */
#define BLOCK_MASK       0x00001fff    /* mask of the page     */
#define BLOCK_MASK_BIT   13            /* mask bit of the page */
#define DATA_T_SIZE      8             /* size of data_t.value */ 

#define CPU_FEATURE_MASK 0x0000000000000107ull /* for   AMASK op */
#define IMPLEMENTATION_VERSION 2               /* for IMPLVER op */

#define FD_MAPPING_MAX 128                   /* syscall read & write */

#define OSF_SIGCONTEXT_SIZE 568 /* size in byte */
#define OSF_FSTAT_SIZE       80 /* size in byte */
#define OSF_RUSAGE_SIZE      72 /* size in byte */

#define OPTION_MAX          30 /* max of options */

/** for systemcall implementation in syscall.cc                 **/
/** I don't implement all, but something used for SPEC CPU2000. **/
#define ASYS_exit                1
#define ASYS_read                3
#define ASYS_write               4
#define ASYS_close               6
#define ASYS_unlink             10
#define ASYS_brk                17
#define	ASYS_lseek	        19
#define ASYS_getpid             20
#define ASYS_getuid             24
#define ASYS_open               45
#define ASYS_getgid             47
#define ASYS_sigprocmask        48
#define ASYS_ioctl              54
#define ASYS_getpagesize        64
#define ASYS_stat               67
#define ASYS_mmap               71
#define ASYS_munmap             73
#define ASYS_fstat              91
#define ASYS_fcntl              92
#define	ASYS_sigreturn         103
#define	ASYS_gettimeofday      116
#define	ASYS_getrusage         117
#define	ASYS_rename            128
#define	ASYS_getrlimit         144
#define	ASYS_setrlimit         145
#define	ASYS_sigaction         156
#define	ASYS_getdirentries     159
#define ASYS_UNIMPLEMENT       226
#define ASYS_osf_getsysinfo    256
#define ASYS_osf_setsysinfo    257

/* Table C-6: Opcode Summary (Alpha Architecture Handbook V.4)   */
#define OP_PAL  0x00  /* PALcode instruction(CALL_PAL)           */
#define OP_LDA  0x08  /* load address instruction                */
#define OP_LDAH 0x09  /* load address high instruction           */
#define OP_INTA 0x10  /* Integer arithmetic instruction          */
#define OP_INTL 0x11  /* Integer logical instruction             */
#define OP_INTS 0x12  /* Integer shift instruction               */
#define OP_INTM 0x13  /* Integer multiply instruction            */
#define OP_ITFP 0x14  /* Integer to floating-point register move */
#define OP_FLTV 0x15  /* VAX floating-point instruction          */
#define OP_FLTI 0x16  /* IEEE floating-point instruction         */
#define OP_FLTL 0x17  /* Floating-point instruction              */
#define OP_MISC 0x18  /* Miscellaneous instruction               */
#define OP_JSR  0x1a  /* Jump instruction                        */
#define OP_FPTI 0x1c  /* Floating-point to integer register move */

#define BIT0    0x0000000000000001ull
#define BIT1    0x0000000000000002ull
#define BIT2    0x0000000000000004ull
#define BIT3    0x0000000000000008ull
#define BIT4    0x0000000000000010ull
#define BIT5    0x0000000000000020ull
#define BIT6    0x0000000000000040ull
#define BIT7    0x0000000000000080ull
#define BIT8    0x0000000000000100ull
#define BIT9    0x0000000000000200ull
#define BIT10   0x0000000000000400ull
#define BIT11   0x0000000000000800ull
#define BIT12   0x0000000000001000ull
#define BIT15   0x0000000000008000ull
#define BIT20   0x0000000000100000ull
#define BIT31   0x0000000080000000ull
#define BIT63   0x8000000000000000ull

#define EXTND8  0xffffffffffffff00ull
#define EXTND16 0xffffffffffff0000ull
#define EXTND21 0xffffffffffe00000ull
#define EXTND32 0xffffffff00000000ull
#define MASK16  0x000000000000ffffull /* mask 16bit */
#define MASK21  0x00000000001fffffull /* mask 21bit */
#define MASK32  0x00000000ffffffffull /* mask 32bit */

#define ZAP0    0xffffffffffffff00ull
#define ZAP1    0xffffffffffff00ffull
#define ZAP2    0xffffffffff00ffffull
#define ZAP3    0xffffffff00ffffffull
#define ZAP4    0xffffff00ffffffffull
#define ZAP5    0xffff00ffffffffffull
#define ZAP6    0xff00ffffffffffffull
#define ZAP7    0x00ffffffffffffffull

/** for instruction.cc **/
#define MSK1 0x3e
#define MSK2 0x3c
#define MSK3 0x38
#define MSK4 0x30


/* 4. structure and class definition                                  */
/**********************************************************************/

struct osf_statbuf{ /*** Used in syscall.c SYS_fstat ***/
  uint32_t st_dev;
  uint32_t st_ino;
  uint32_t st_mode;
  uint16_t st_nlink;
  uint16_t pad0;
  uint32_t st_uid;
  uint32_t st_gid;
  uint32_t st_rdev;
  uint32_t pad1;
  uint64_t st_size;
  uint32_t st_atime;
  uint32_t st_spare1;
  uint32_t st_mtime;
  uint32_t st_spare2;
  uint32_t st_ctime;
  uint32_t st_spare3;
  uint32_t st_blksize;
  uint32_t st_blocks;
  uint32_t st_gennum;
  uint32_t st_spare4;
};

class system_config{
 public:
  char program_name[512];
  uint64_t end_tc;
  int verbose_interval;
  int debug_flag;
  int syscall_flag;
  int fd_adjust;
  int fd_stdin_flag;
  int fd_stdin; /* file descripter for STDIN */
  system_config(char *, char**);
};


class evaluation_result{
 public:
  uint64_t retired_inst; 
  int used_memory_block; 
  time_t time_begin;     /* start time stamp */
  struct timeval tp;     /* start time stamp */
  struct timezone tzp;   /* start time stamp */
  evaluation_result();
};


class data_t{
  uint64_t value;
 public:
  int cmov;      
  uint64_t ld();     
  int   st(uint64_t);
  int init(uint64_t);
};


class architecture_state{
 public:
  data_t pc;    /* program counter      */
  data_t r[32]; /* general purpose regs */
  data_t f[32]; /* floating point  regs */
  uint64_t rr(int);/* Register Read     */
  architecture_state(system_config *,
		     evaluation_result *);
};


class main_memory{
  evaluation_result *e;
  data_t *block_table[BLOCK_TABLE_SIZE];
  data_t *allocblock(data_t *);
 public:
  void ld_8byte(data_t *, data_t *);
  void st_8byte(data_t *, data_t *, DATA_TYPE);
  main_memory(evaluation_result *);
};


class memory_system{
  evaluation_result *e;
  class main_memory *mm;
  void ld_8byte(data_t *, data_t *);
  void st_8byte(data_t *, data_t *, DATA_TYPE);
 public:
  void ld_inst(data_t *, INST_TYPE *);
  void ld_nbyte(int, data_t *, data_t *);
  void st_nbyte(int, data_t *, data_t *);
  ~memory_system();
  memory_system(system_config *, 
		evaluation_result *);
};


class system_manager{
  class evaluation_result *e;
  class system_config     *ms;
  architecture_state      *as;
  memory_system           *mem;
  /* File Descriptor Mapping Table */
  DATA_TYPE fd_table[FD_MAPPING_MAX]; 
  /* Unique Data for rduniq & wruniq inst */
  DATA_TYPE uniq; 
 public:
  int running;    /* chip is running ? */
  void execute_pal(class instruction *);
  int translate_fd(DATA_TYPE);
  system_manager(architecture_state *, 
		 memory_system *,
		 system_config *,
		 evaluation_result *);
};


class debug{
  class evaluation_result *e;
  class system_config     *ms;
  architecture_state      *as;
  memory_system           *mem;
 public:
  debug(architecture_state *, memory_system *,
	struct system_config *, evaluation_result *);
  void debugmode();
};


class instruction{
  evaluation_result  *e;
  architecture_state *as;
  system_manager     *sys;
  memory_system      *mem;

  INST_TYPE ir; /* 32bit instruction code  */
  int Op;       /* Opcode field            */
  int RA;       /* Ra field of the inst    */
  int RB;       /* Rb field of the inst    */
  int RC;       /* Rc field of the inst    */
  int ST;       /* store inst ?            */
  int LD;       /* load inst ?             */
  int CM;       /* CMOV inst ?             */
  int LA;       /* load address inst ?     */
  int BR;       /* branch inst ?           */
  int Ai;       /* Rav is immediate ?      */
  int Bi;       /* Rbv is immediate ?      */
  int Af;       /* Rav from floating-reg ? */
  int Bf;       /* Rbv from floating-reg ? */
  int WF;       /* Write to the f-reg ?    */
  int WB;       /* Writeback reg index     */
  data_t  Cpc;  /* PC of the instruction   */
  data_t  Npc;  /* PC of the next inst.    */
  data_t  Imm;  /* immediate               */
  data_t  Adr;  /* load & store address    */
  data_t  Rav;  /* Ra                      */
  data_t  Rbv;  /* Rb                      */
  data_t  Rcv;  /* Rc                      */
 public:
  int Fetch(data_t *);
  int Fetch(data_t *, INST_TYPE);
  int Slot();
  int Rename();
  int Issue();
  int RegisterRead();
  int Execute(data_t *);
  int Memory();
  int WriteBack();

  INST_TYPE get_ir();
  int data_ld(data_t *, data_t *);
  int data_st(data_t *, data_t *);
  instruction(architecture_state *,
              memory_system *,
              system_manager *,
              system_config *,
              evaluation_result *);
};


class simple_chip{
  system_config      *sc;
  evaluation_result  *e;
  debug              *deb;
  system_manager     *sys;
  instruction        *p;
 public:
  memory_system      *mem;
  architecture_state *as;
  simple_chip(char *, char **);
  ~simple_chip();
  int step();
};


struct alpha_stat{  /*** for syscall.cc ***/
  uint64_t  st_dev;         
  uint32_t  st_ino;              
   int32_t  __pad1;
  uint32_t  st_mode;            
  uint32_t  st_nlink;          
  uint32_t  st_uid;              
  uint32_t  st_gid;              
  uint64_t  st_rdev;             
   int64_t  st_size;             
   int64_t  st_atime;           
   int64_t  st_mtime;           
   int64_t  st_ctime;
  uint32_t  st_blocks;        
   int32_t  __pad2;
  uint32_t  st_blksize;     
  uint32_t  st_flags;
  uint32_t  st_gen;
   int32_t  __pad3;
   int64_t  __unused[4];
};


/* 5. function prototypes                                             */
/**********************************************************************/
extern void print_evaluation_result(struct evaluation_result *,
				    struct system_config *,
				    system_manager *);
extern void usage();
extern uint64_t map_s(unsigned int);
extern INST_TYPE cmov_ir_create(INST_TYPE);
extern int ALU(INST_TYPE, data_t *, data_t *, data_t *);
extern int BRU(INST_TYPE, data_t *, data_t *, data_t *, data_t *);
extern void option_from_file(char **, char *);

extern int execute_cmovb(instruction *, architecture_state *);
extern void house_keeper(system_manager *, system_config *,
			 evaluation_result *, debug *);
extern "C" int simple_cpu(int, uint64_t *); /**** C interface *****/
